from typing import Optional

import numpy as np

from autofit.graphical.expectation_propagation import AbstractFactorOptimiser
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.prior_model import AbstractPriorModel
from autofit.non_linear.analysis import Analysis
from autofit.non_linear.paths.abstract import AbstractPaths
from .abstract import AbstractModelFactor


class AnalysisFactor(AbstractModelFactor):
    @property
    def prior_model(self):
        return self._prior_model

    def __init__(
            self,
            prior_model: AbstractPriorModel,
            analysis: Analysis,
            optimiser: Optional[AbstractFactorOptimiser] = None,
            name=None
    ):
        """
        A factor in the graph that actually computes the likelihood of a model
        given values for each variable that model contains

        Parameters
        ----------
        prior_model
            A model with some dimensionality
        analysis
            A class that implements a function which evaluates how well an
            instance of the model fits some data
        optimiser
            A custom optimiser that will be used to fit this factor specifically
            instead of the default optimiser
        """
        self.label = prior_model.label
        self.analysis = analysis

        def _factor(
                **kwargs: np.ndarray
        ) -> float:
            """
            Returns an instance of the prior model and evaluates it, forming
            a factor.

            Parameters
            ----------
            kwargs
                Arguments with names that are unique for each prior.

            Returns
            -------
            Calculated likelihood
            """
            arguments = dict()
            for name_, array in kwargs.items():
                prior_id = int(name_.split("_")[1])
                prior = prior_model.prior_with_id(
                    prior_id
                )
                arguments[prior] = array
            instance = prior_model.instance_for_arguments(
                arguments
            )
            return analysis.log_likelihood_function(
                instance
            )

        prior_variable_dict = {
            prior.name: prior
            for prior
            in prior_model.priors
        }

        super().__init__(
            prior_model=prior_model,
            factor=_factor,
            optimiser=optimiser,
            prior_variable_dict=prior_variable_dict,
            name=name
        )

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, item):
        return getattr(self.prior_model, item)

    def name_for_variable(
            self,
            variable
    ):
        path = ".".join(
            self.prior_model.path_for_prior(
                variable
            )
        )
        return f"{self.name}.{path}"

    def visualize(
            self,
            paths: AbstractPaths,
            instance: ModelInstance,
            during_analysis: bool
    ):
        """
        Visualise the instances provided using each factor.

        Instances in the ModelInstance must have the same order as the factors.

        Parameters
        ----------
        paths
            Object describing where data should be saved to
        instance
            A collection of instances, each corresponding to a factor
        during_analysis
            Is this visualisation during analysis?
        """
        self.analysis.visualize(
            paths,
            instance,
            during_analysis
        )

    def log_likelihood_function(
            self,
            instance: ModelInstance
    ) -> float:
        return self.analysis.log_likelihood_function(instance)



import os

from autofit.mapper.prior_model.collection import Collection

path = os.path.dirname(os.path.realpath(__file__))


class ModelMapper(Collection):
    """
    A mapper of priors formed by passing in classes to be reconstructed

    @DynamicAttrs

    The ModelMapper converts a set of classes whose input attributes may be
    modeled using a non-linear search, to parameters with priors attached.

    A config is passed into the model mapper to provide default setup values for
    the priors:

    mapper = ModelMapper(config)

    All class instances that are to be generated by the model mapper are
    specified by adding classes to it:

    mapper = ModelMapper()

    mapper.sersic = al.lp.AbstractSersic
    mapper.gaussian = al.lp.Gaussian
    mapper.any_class = SomeClass

    A `Model` instance is created each time we add a class to the mapper. We
    can access those models using # the mapper attributes:

    sersic_model = mapper.sersic

    This allows us to replace the default priors:

    mapper.sersic.normalization = GaussianPrior(mean=2., sigma=5.)

    Or maybe we want to tie two priors together:

    mapper.sersic.two = mapper.other_sersic.two

    This statement reduces the number of priors by one and means that the two
    sersic instances will always share # the same rotation two two.

    We can then create instances of every class for a unit hypercube vector
    with length equal to # len(mapper.priors):

    model_instance = mapper.model_instance_for_vector([.4, .2, .3, .1])

    The attributes of the model_instance are named the same as those of the mapper:

    sersic_1 = mapper.sersic_1

    But this attribute is an instance of the actual AbstractSersic:P
    class

    A ModelMapper can be concisely constructed using keyword arguments:

    mapper = prior.ModelMapper(
        source_light_profile=light_profile.AbstractSersic,
        lens_mass_profile=mass_profile.IsothermalCore,
        lens_light_profile=light_profile.SersicCore
    )
    """

    @property
    def prior_prior_model_dict(self):
        """

        Returns
        -------
        prior_prior_model_dict: {Prior: Model}
            A dictionary mapping priors to associated prior models. Each prior will only
            have one prior model; if a prior is shared by two prior models then one of
            those prior models will be in this dictionary.
        """
        return {
            prior: prior_model[1]
            for prior_model in self.prior_model_tuples
            for _, prior in prior_model[1].prior_tuples
        }



from autofit.plot import SamplesPlotter

from autofit.plot.samples_plotters import skip_plot_in_test_mode
from autofit.plot.samples_plotters import log_value_error

class UltraNestPlotter(SamplesPlotter):

    @skip_plot_in_test_mode
    @log_value_error
    def cornerplot(self, **kwargs):
        """
        Plots the in-built ``ultranest`` plot ``cornerplot``.

        This figure plots a corner plot of the 1-D and 2-D marginalized posteriors.
        """

        from ultranest import plot

        plot.cornerplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="cornerplot")
        self.close()

    @skip_plot_in_test_mode
    @log_value_error
    def runplot(self, **kwargs):
        """
        Plots the in-built ``ultranest`` plot ``runplot``.

        This figure plots live points, ln(likelihood), ln(weight), and ln(evidence) vs. ln(prior volume).
        """
        from ultranest import plot

        plot.runplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="runplot")
        self.close()

    @skip_plot_in_test_mode
    @log_value_error
    def traceplot(self, **kwargs):
        """
        Plots the in-built ``ultranest`` plot ``traceplot``.

        This figure plots traces and marginalized posteriors for each parameter.
        """
        from ultranest import plot

        plot.traceplot(
            results=self.samples.results_internal,
            **kwargs
        )

        self.output.to_figure(structure=None, auto_filename="traceplot")
        self.close()


import logging
import multiprocessing
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Optional, List, Tuple

from autofit import exc
from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.expectation_propagation.history import EPHistory
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.graphical.mean_field import Status, MeanField, FactorApproximation
from autofit.graphical.utils import StatusFlag, LogWarnings
from autofit.mapper.identifier import Identifier
from autofit.non_linear.paths import DirectoryPaths
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.tools.util import IntervalCounter
from .factor_optimiser import AbstractFactorOptimiser, ExactFactorFit
from .visualise import Visualise

logger = logging.getLogger(__name__)


class ApproxUpdater(ABC):
    """
    Handles updating of the EPMeanField using new model distributions from
    individual factors
    """

    @abstractmethod
    def delta(self, factor: Factor, model_approx: EPMeanField):
        """
        Compute a delta for a given factor. This dictates how much the message
        for that factor is updated. A delta of 0 is no update and a delta of 1
        is maximum update.
        """

    def update_model_approx(
        self,
        new_model_dist: MeanField,
        factor_approx: FactorApproximation,
        model_approx: EPMeanField,
        status: Status = Status(),
    ) -> Tuple[EPMeanField, Status]:
        delta = self.delta(factor=factor_approx.factor, model_approx=model_approx,)

        return model_approx.project_mean_field(
            new_model_dist, factor_approx, delta=delta, status=status,
        )


class SimplerUpdater(ApproxUpdater):
    def __init__(self, delta: float = 1.0):
        """
        Simply set delta to a fixed value

        Parameters
        ----------
        delta
            A fixed rate at which all factors update
        """
        self._delta = delta

    def delta(self, factor, model_approx):
        return self._delta


class FactorUpdater(SimplerUpdater):
    def __init__(self, factor_deltas: Dict[Factor, float], default=1.0):
        """
        Determine how fast factors update on a factor by factor basis.

        Parameters
        ----------
        factor_deltas
            A dictionary dictating how fast each factor should update the mean field.
        default
            A default value for when no value is provided.
        """
        super().__init__(delta=default)
        self.factor_deltas = factor_deltas

    def delta(self, factor, model_approx):
        return self.factor_deltas.get(factor, self._delta)


class DynamicUpdater(SimplerUpdater):
    def delta(self, factor, model_approx):
        """
        Variables are updated dynamically. The more factors that share a variable
        the slower that variable is updated. This accounts for cases when a variable
        is shared by many more factors than another variable and so shrinks too
        quickly.
        """

        variable_message_count = model_approx.variable_message_count
        min_value = min(variable_message_count.values())
        return MeanField(
            {
                variable: self._delta * (min_value / message_count)
                for variable, message_count in variable_message_count.items()
            }
        )


def factor_step(factor_approx, optimiser):
    factor = factor_approx.factor
    factor_logger = logging.getLogger(factor.name)
    factor_logger.debug("Optimising...")
    try:
        with LogWarnings(
            logger=factor_logger.debug, action="always"
        ) as caught_warnings:
            new_model_dist, status = optimiser.optimise(factor_approx)

        messages = status.messages + tuple(caught_warnings.messages)

        status = Status(status.success, messages, status.flag, result=status.result)

    except (ValueError, ArithmeticError, RuntimeError) as e:
        logger.exception(e)
        status = Status(
            False, (f"Factor: {factor} experienced error {e}",), StatusFlag.FAILURE,
        )
        new_model_dist = factor_approx.model_dist

    factor_logger.debug(status)
    return new_model_dist, status


class EPOptimiser:
    def __init__(
        self,
        factor_graph: FactorGraph,
        default_optimiser: Optional[AbstractFactorOptimiser] = None,
        factor_optimisers: Optional[Dict[Factor, AbstractFactorOptimiser]] = None,
        ep_history: Optional[EPHistory] = None,
        factor_order: Optional[List[Factor]] = None,
        paths: AbstractPaths = None,
        updater: Optional[ApproxUpdater] = None,
    ):
        """
        Optimise a factor graph.

        Cycles through factors optimising them individually using priors
        created through expectation propagation; in effect the prior for each
        variable for a given factor is the product of the posteriors of that
        variable for all other factors.

        Parameters
        ----------
        factor_graph
            A graph describing the relationships between multiple factors
        default_optimiser
            An optimiser that is used if no specific optimiser is provided for a factor
        factor_optimisers
            Specific optimisers used to optimise each factor
        ep_history
            Optionally specify an alternate history
        factor_order
            The factors in the graph but placed in the order in which they should
            be optimised
        paths
            Optionally define how data should be output
        """
        self.paths = paths or DirectoryPaths(identifier=str(Identifier(factor_graph)))

        factor_optimisers = factor_optimisers or {}
        self.factor_graph = factor_graph
        self.factors = factor_order or self.factor_graph.factors
        self.default_optimiser = default_optimiser
        self.updater = updater or SimplerUpdater(delta=1.0)

        if default_optimiser is None:
            self.factor_optimisers = factor_optimisers
            missing = set(self.factors) - self.factor_optimisers.keys()
            if missing:
                raise (
                    ValueError(
                        f"missing optimisers for {missing}, "
                        "pass a default_optimiser or add missing optimsers"
                    )
                )
        else:
            self.factor_optimisers = {
                factor: factor_optimisers.get(factor, default_optimiser)
                for factor in self.factors
            }

        for optimiser in self.factor_optimisers.values():
            optimiser.paths = self.paths

        self.ep_history = ep_history or EPHistory()

        with open(self.output_path / "graph.info", "w+") as f:
            f.write(self.factor_graph.info)

        self.visualiser = Visualise(self.ep_history, self.output_path)

    @classmethod
    def from_meanfield(
        cls,
        model_approx: EPMeanField,
        default_optimiser: Optional[AbstractFactorOptimiser] = None,
        factor_optimisers: Optional[Dict[Factor, AbstractFactorOptimiser]] = None,
        ep_history: Optional[EPHistory] = None,
        factor_order: Optional[List[Factor]] = None,
        paths: AbstractPaths = None,
        exact_fit_kws=None,
    ):
        factor_graph = model_approx.factor_graph
        factor_order = factor_order or factor_graph.factors

        factor_optimisers = factor_optimisers or {}
        factor_mean_field = model_approx.factor_mean_field
        for factor in factor_graph.factors:
            if factor not in factor_optimisers:
                factor_dist = factor_mean_field[factor]
                if factor.has_exact_projection(factor_dist):
                    factor_optimisers[factor] = ExactFactorFit(**(exact_fit_kws or {}))
                else:
                    factor_optimisers[factor] = default_optimiser

        return cls(
            factor_graph=factor_graph,
            default_optimiser=default_optimiser,
            factor_optimisers=factor_optimisers,
            ep_history=ep_history,
            factor_order=factor_order,
            paths=paths,
        )

    @property
    def output_path(self) -> Path:
        """
        The path at which data will be output. Uses the name of the optimiser.

        If the path does not exist it is created.
        """
        path = Path(self.paths.output_path)
        os.makedirs(path, exist_ok=True)
        return path

    def _log_factor(self, factor: Factor):
        """
        Log information for the factor and its history.
        """
        factor_logger = logging.getLogger(factor.name)
        try:
            factor_history = self.ep_history[factor]
            if factor_history.history:
                log_evidence = factor_history.latest_update.log_evidence
                divergence = factor_history.kl_divergence()

                factor_logger.info(f"Log Evidence = {log_evidence}")
                factor_logger.info(f"KL Divergence = {divergence}")
        except exc.HistoryException as e:
            factor_logger.exception(e)

    def factor_step(self, factor_approx, optimiser):
        return factor_step(factor_approx, optimiser)

    def run(
        self,
        model_approx: EPMeanField,
        max_steps: int = 100,
        log_interval: int = 10,
        visualise_interval: int = 100,
        output_interval: int = 10,
    ) -> EPMeanField:
        """
        Run the optimisation on an approximation of the model.

        Parameters
        ----------
        model_approx
            A collection of messages describing priors on the model's variables.
        max_steps
            The maximum number of steps prior to termination. Termination may also
            occur when difference in log evidence or KL Divergence drop below a given
            threshold for two consecutive optimisations of a given factor.
        log_interval
            How steps should we wait before logging information?
        visualise_interval
            How steps should we wait before visualising information?
            This includes plots of KL Divergence and Evidence.
        output_interval
            How steps should we wait before outputting information?
            This includes the model.results file which describes the current mean values
            of each message.

        Returns
        -------
        An updated approximation of the model
        """
        should_log = IntervalCounter(log_interval)
        should_visualise = IntervalCounter(visualise_interval)
        should_output = IntervalCounter(output_interval)

        for _ in range(max_steps):
            _should_log = should_log()
            _should_visualise = should_visualise()
            _should_output = should_output()
            for factor, optimiser in self.factor_optimisers.items():
                factor_approx = model_approx.factor_approximation(factor)
                new_model_dist, status = self.factor_step(factor_approx, optimiser)
                model_approx, status = self.updater.update_model_approx(
                    new_model_dist, factor_approx, model_approx, status
                )
                if status and _should_log:
                    self._log_factor(factor)

                if self.ep_history(factor, model_approx, status):
                    logger.info("Terminating optimisation")
                    break  # callback controls convergence

            else:  # If no break do next iteration
                if _should_visualise:
                    self.visualiser()
                if _should_output:
                    self._output_results(model_approx)
                continue
            break  # stop iterations

        self.visualiser()
        self._output_results(model_approx)

        return model_approx

    def _output_results(self, model_approx: EPMeanField):
        """
        Save the graph.results text
        """
        with open(self.output_path / "graph.results", "w+") as f:
            f.write(self.factor_graph.make_results_text(model_approx))


class ParallelEPOptimiser(EPOptimiser):
    def __init__(
        self,
        factor_graph: FactorGraph,
        n_cores: int,
        default_optimiser: Optional[AbstractFactorOptimiser] = None,
        factor_optimisers: Optional[Dict[Factor, AbstractFactorOptimiser]] = None,
        ep_history: Optional[EPHistory] = None,
        factor_order: Optional[List[Factor]] = None,
        paths: AbstractPaths = None,
        updater: Optional[ApproxUpdater] = None,
    ):
        """
        Optimise a factor graph.

        Optimises all factors simultaneously on parallel processes, combines the results
        and repeats.

        Parameters
        ----------
        factor_graph
            A graph describing the relationships between multiple factors
        n_cores
            How many cores are available? The main process takes one core so the
            multiprocessing pool has n_cores - 1 processes. Should be at least 3
        default_optimiser
            An optimiser that is used if no specific optimiser is provided for a factor
        factor_optimisers
            Specific optimisers used to optimise each factor
        ep_history
            Optionally specify an alternate history
        factor_order
            The factors in the graph but placed in the order in which they should
            be optimised
        paths
            Optionally define how data should be output
        """

        if n_cores < 3:
            raise AssertionError(
                "With less than three cores it is better to use EPOptimiser"
            )

        super().__init__(
            factor_graph=factor_graph,
            default_optimiser=default_optimiser,
            factor_optimisers=factor_optimisers,
            ep_history=ep_history,
            factor_order=factor_order,
            paths=paths,
            updater=updater,
        )
        self.pool = multiprocessing.Pool(n_cores - 1)

    def run(
        self,
        model_approx: EPMeanField,
        max_steps: int = 100,
        log_interval: int = 10,
        visualise_interval: int = 100,
        output_interval: int = 10,
    ) -> EPMeanField:
        """
        Run the optimisation on an approximation of the model.

        Parameters
        ----------
        model_approx
            A collection of messages describing priors on the model's variables.
        max_steps
            The maximum number of steps prior to termination. Termination may also
            occur when difference in log evidence or KL Divergence drop below a given
            threshold for two consecutive optimisations of a given factor.
        log_interval
            How steps should we wait before logging information?
        visualise_interval
            How steps should we wait before visualising information?
            This includes plots of KL Divergence and Evidence.
        output_interval
            How steps should we wait before outputting information?
            This includes the model.results file which describes the current mean values
            of each message.

        Returns
        -------
        An updated approximation of the model
        """
        should_log = IntervalCounter(log_interval)
        should_visualise = IntervalCounter(visualise_interval)
        should_output = IntervalCounter(output_interval)

        for _ in range(max_steps):
            _should_log = should_log()
            _should_visualise = should_visualise()
            _should_output = should_output()

            factor_approx_optimisers = [
                (model_approx.factor_approximation(factor), optimiser)
                for factor, optimiser in self.factor_optimisers.items()
            ]

            new_dist_statuses = self.pool.starmap(factor_step, factor_approx_optimisers)

            for (factor_approx, _), (new_model_dist, status) in zip(
                factor_approx_optimisers, new_dist_statuses
            ):
                model_approx, status = self.updater.update_model_approx(
                    new_model_dist, factor_approx, model_approx, status
                )
                factor = factor_approx.factor
                if status and _should_log:
                    self._log_factor(factor)

                if self.ep_history(factor, model_approx, status):
                    logger.info("Terminating optimisation")
                    break  # callback controls convergence

            else:  # If no break do next iteration
                if _should_visualise:
                    self.visualiser()
                if _should_output:
                    self._output_results(model_approx)
                continue
            break  # stop iterations

        self.visualiser()
        self._output_results(model_approx)

        return model_approx



import builtins
import copy
import inspect
import logging

from autoconf.class_path import get_class_path
from autoconf.exc import ConfigException
from autofit.mapper.model import assert_not_frozen
from autofit.mapper.model_object import ModelObject
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior.deferred import DeferredInstance
from autofit.mapper.prior.tuple_prior import TuplePrior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.abstract import check_assertions
from autofit.tools.namer import namer

logger = logging.getLogger(__name__)

class_args_dict = dict()


class Model(AbstractPriorModel):
    """
    @DynamicAttrs
    """

    @property
    def name(self):
        return self.cls.__name__

    def __str__(self):
        prior_string = ", ".join(map(str, self.prior_tuples))
        return f"{self.name} {prior_string}"

    def __repr__(self):
        return f"<{self.__class__.__name__} {self}>"

    def as_model(self):
        return Model(self.cls)

    def __hash__(self):
        return self.id

    def __add__(self, other):
        if self.cls != other.cls:
            raise TypeError(
                f"Cannot add PriorModels with different classes "
                f"({self.cls.__name__} and {other.cls.__name__})"
            )
        return super().__add__(other)

    def __init__(self, cls, **kwargs):
        """
        The object a Python class is input into to create a model-component, which has free parameters that are fitted
        by a non-linear search.

        The ``Model`` object is flexible, and can create models from many input Python data structures
        (e.g. a list of classes, dictionary of classes, hierarchy of classes).

        For a complete description of the model composition API, see the **PyAutoFit** model API cookbooks:

        https://pyautofit.readthedocs.io/en/latest/cookbooks/cookbook_1_basics.html

        The Python class input into a ``Model`` to create a model component is written using the following format:

        - The name of the class is the name of the model component (e.g. ``Gaussian``).
        - The input arguments of the constructor are the parameters of the mode (e.g. ``centre``, ``normalization`` and ``sigma``).
        - The default values of the input arguments tell PyAutoFit whether a parameter is a single-valued float or a
        multi-valued tuple.

        [Rich explain everything else]

        Parameters
        ----------
        cls
            The class associated with this instance

        Examples
        --------

        class Gaussian:

            def __init__(
                self,
                centre=0.0,        # <- PyAutoFit recognises these
                normalization=0.1, # <- constructor arguments are
                sigma=0.01,        # <- the Gaussian's parameters.
            ):
                self.centre = centre
                self.normalization = normalization
                self.sigma = sigma

        model = af.Model(Gaussian)
        """
        super().__init__(label=namer(cls.__name__) if inspect.isclass(cls) else None)
        if cls is self:
            return

        if not (inspect.isclass(cls) or inspect.isfunction(cls)):
            raise AssertionError(f"{cls} is not a class or function")

        self.cls = cls

        try:
            annotations = inspect.getfullargspec(cls).annotations
            for key, value in annotations.items():
                if isinstance(value, str):
                    annotations[key] = getattr(builtins, value)
        except TypeError:
            annotations = dict()

        try:
            arg_spec = inspect.getfullargspec(cls)
            defaults = dict(
                zip(arg_spec.args[-len(arg_spec.defaults) :], arg_spec.defaults)
            )
        except TypeError:
            defaults = {}

        args = self.constructor_argument_names

        if "settings" in defaults:
            del defaults["settings"]
        if "settings" in args:
            args.remove("settings")

        for arg in args:
            if isinstance(defaults.get(arg), str):
                continue

            if arg in kwargs:
                keyword_arg = kwargs[arg]
                if isinstance(keyword_arg, (list, dict)):
                    from autofit.mapper.prior_model.collection import Collection

                    ls = Collection(keyword_arg)

                    setattr(self, arg, ls)
                else:
                    if inspect.isclass(keyword_arg):
                        keyword_arg = Model(keyword_arg)
                    setattr(self, arg, keyword_arg)
            elif arg in defaults and isinstance(defaults[arg], tuple):
                tuple_prior = TuplePrior()
                for i in range(len(defaults[arg])):
                    attribute_name = "{}_{}".format(arg, i)
                    setattr(
                        tuple_prior, attribute_name, self.make_prior(attribute_name)
                    )
                setattr(self, arg, tuple_prior)
            elif arg in annotations and annotations[arg] is not float:
                spec = annotations[arg]

                # noinspection PyUnresolvedReferences
                if inspect.isclass(spec) and issubclass(spec, float):
                    from autofit.mapper.prior_model.annotation import (
                        AnnotationPriorModel,
                    )

                    setattr(self, arg, AnnotationPriorModel(spec, cls, arg))
                elif hasattr(spec, "__args__") and type(None) in spec.__args__:
                    setattr(self, arg, None)
                else:
                    setattr(self, arg, Model(annotations[arg]))
            else:
                prior = self.make_prior(arg)
                if (
                    isinstance(prior, ConfigException)
                    and hasattr(cls, "__default_fields__")
                    and arg in cls.__default_fields__
                ):
                    prior = defaults[arg]
                setattr(self, arg, prior)
        for key, value in kwargs.items():
            if not hasattr(self, key):
                setattr(self, key, Model(value) if inspect.isclass(value) else value)

    def dict(self):
        return {"class_path": get_class_path(self.cls), **super().dict()}

    # noinspection PyAttributeOutsideInit
    @property
    def constructor_argument_names(self):
        if self.cls not in class_args_dict:
            try:
                class_args_dict[self.cls] = inspect.getfullargspec(self.cls).args[1:]
            except TypeError:
                class_args_dict[self.cls] = []
        return class_args_dict[self.cls]

    def __eq__(self, other):
        return (
            isinstance(other, Model)
            and self.cls == other.cls
            and self.prior_tuples == other.prior_tuples
        )

    def make_prior(self, attribute_name):
        """
        Returns a prior for an attribute of a class with a given name. The prior is
        created by searching the default prior config for the attribute.

        Entries in configuration with a u become uniform priors; with a g become
        gaussian priors; with a c become instances.

        If prior configuration for a given attribute is not specified in the
        configuration for a class then the configuration corresponding to the parents
        of that class is searched. If no configuration can be found then a prior
        exception is raised.

        Parameters
        ----------
        attribute_name: str
            The name of the attribute for which a prior is created

        Returns
        -------
        prior: p.Prior
            A prior

        Raises
        ------
        exc.PriorException
            If no configuration can be found
        """
        cls = self.cls
        if not inspect.isclass(cls):
            # noinspection PyProtectedMember
            cls = inspect._findclass(cls)
        try:
            return Prior.for_class_and_attribute_name(cls, attribute_name)
        except ConfigException as e:
            return e

    @assert_not_frozen
    def __setattr__(self, key, value):
        try:
            value.label = namer(key)
        except (AttributeError, TypeError):
            pass

        if key not in (
            "component_number",
            "phase_property_position",
            "mapping_name",
            "id",
            "_is_frozen",
            "_frozen_cache",
        ):
            try:
                if "_" in key:
                    name = key.split("_")[0]
                    tuple_prior = [v for k, v in self.tuple_prior_tuples if name == k][
                        0
                    ]
                    setattr(tuple_prior, key, value)
                    return
            except IndexError:
                pass
        try:
            super().__setattr__(key, value)
        except AttributeError as e:
            logger.exception(e)
            logger.exception(key)

    def __getattr__(self, item):
        try:
            if (
                "_" in item
                and item not in ("_is_frozen", "tuple_prior_tuples")
                and not item.startswith("_")
            ):
                return getattr(
                    [v for k, v in self.tuple_prior_tuples if item.split("_")[0] == k][
                        0
                    ],
                    item,
                )

        except IndexError:
            pass

        self.__getattribute__(item)

    @property
    def is_deferred_arguments(self):
        return len(self.direct_deferred_tuples) > 0

    # noinspection PyUnresolvedReferences
    @check_assertions
    def _instance_for_arguments(self, arguments: {ModelObject: object}):
        """
        Returns an instance of the associated class for a set of arguments

        Parameters
        ----------
        arguments: {Prior: float}
            Dictionary mapping_matrix priors to attribute analysis_path and value pairs

        Returns
        -------
            An instance of the class
        """
        model_arguments = dict()
        attribute_arguments = {
            key: value
            for key, value in self.__dict__.items()
            if key in self.constructor_argument_names
        }

        for tuple_prior in self.tuple_prior_tuples:
            model_arguments[tuple_prior.name] = tuple_prior.prior.value_for_arguments(
                arguments
            )
        for prior_model_tuple in self.direct_prior_model_tuples:
            prior_model = prior_model_tuple.prior_model
            model_arguments[
                prior_model_tuple.name
            ] = prior_model.instance_for_arguments(arguments,)

        prior_arguments = dict()

        for name, prior in self.direct_prior_tuples:
            try:
                prior_arguments[name] = arguments[prior]
            except KeyError as e:
                raise KeyError(f"No argument given for prior {name}") from e

        constructor_arguments = {
            **attribute_arguments,
            **model_arguments,
            **prior_arguments,
        }

        if self.is_deferred_arguments:
            return DeferredInstance(self.cls, constructor_arguments)

        if not inspect.isclass(self.cls):
            result = object.__new__(inspect._findclass(self.cls))
            cls = self.cls
            cls(result, **constructor_arguments)
        else:
            result = self.cls(**constructor_arguments)

        for key, value in self.__dict__.items():
            if (
                not hasattr(result, key)
                and not isinstance(value, Prior)
                and not key == "cls"
                and not key.startswith("_")
            ):
                if isinstance(value, Model):
                    value = value.instance_for_arguments(arguments)
                elif isinstance(value, Prior):
                    value = arguments[value]
                try:
                    setattr(result, key, value)
                except AttributeError:
                    pass

        return result

    def gaussian_prior_model_for_arguments(self, arguments):
        """
        Returns a new instance of model mapper with a set of Gaussian priors based on \
        tuples provided by a previous nonlinear search.

        Parameters
        ----------
        arguments: [(float, float)]
            Tuples providing the mean and sigma of gaussians

        Returns
        -------
        new_model: ModelMapper
            A new model mapper populated with Gaussian priors
        """
        self.unfreeze()
        new_model = copy.deepcopy(self)

        new_model._assertions = list()

        model_arguments = {t.name: arguments[t.prior] for t in self.direct_prior_tuples}

        for tuple_prior_tuple in self.tuple_prior_tuples:
            setattr(
                new_model,
                tuple_prior_tuple.name,
                tuple_prior_tuple.prior.gaussian_tuple_prior_for_arguments(arguments),
            )
        for prior_tuple in self.direct_prior_tuples:
            setattr(new_model, prior_tuple.name, model_arguments[prior_tuple.name])
        for instance_tuple in self.direct_instance_tuples:
            setattr(new_model, instance_tuple.name, instance_tuple.instance)

        for name, prior_model in self.direct_prior_model_tuples:
            setattr(
                new_model,
                name,
                prior_model.gaussian_prior_model_for_arguments(arguments),
            )

        return new_model



from .factor import AbstractFactor, Factor
from .graph import FactorGraph
from .transform import (
    VariableTransform,
    FullCholeskyTransform,
    identity_transform,
    TransformedNode,
)
from ...mapper.variable import Variable, Plate, FactorValue

dir(FactorValue)



from autofit.messages.normal import NormalMessage
from .abstract import Prior
from ...messages.composed_transform import TransformedMessage
from ...messages.transform import log_transform


class LogGaussianPrior(Prior):

    __identifier_fields__ = ("lower_limit", "upper_limit", "mean", "sigma")

    def __init__(
        self,
        mean : float,
        sigma : float,
        lower_limit : float = 0.0,
        upper_limit : float = float("inf"),
        id_=None,
    ):
        """
        A prior with a log base 10 uniform distribution, defined between a lower limit and upper limit.

        The conversion of an input unit value, ``u``, to a physical value, ``p``, via the prior is as follows:

        .. math::

            p = \mu + (\sigma * sqrt(2) * erfcinv(2.0 * (1.0 - u))

        For example for ``prior = LogGaussianPrior(mean=1.0, sigma=2.0)``, an
        input ``prior.value_for(unit=0.5)`` is equal to 1.0.

        [Rich describe how this is done via message]

        Parameters
        ----------
        mean
            The mean of the Gaussian distribution defining the prior.
        sigma
            The sigma value of the Gaussian distribution defining the prior.
        lower_limit
            A lower limit of the Gaussian distribution; physical values below this value are rejected.
        upper_limit
            A upper limit of the Gaussian distribution; physical values below this value are rejected.

        Examples
        --------

        prior = af.LogGaussianPrior(mean=1.0, sigma=2.0, lower_limit=0.0, upper_limit=2.0)

        physical_value = prior.value_for(unit=0.5)
        """
        lower_limit = float(lower_limit)
        upper_limit = float(upper_limit)

        self.mean = mean
        self.sigma = sigma

        message = TransformedMessage(NormalMessage(mean, sigma), log_transform,)

        super().__init__(
            message=message, lower_limit=lower_limit, upper_limit=upper_limit, id_=id_,
        )

    def _new_for_base_message(self, message):
        """
        Create a new instance of this wrapper but change the parameters used
        to instantiate the underlying message. This is useful for retaining
        the same transform stack after recreating the underlying message during
        projection.
        """
        return LogGaussianPrior(
            *message.parameters,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
            id_=self.instance().id,
        )

    def value_for(self, unit : float, ignore_prior_limits : bool = False) -> float:
        """
        Return a physical value for a value between 0 and 1 with the transformation
        described by this prior.

        Parameters
        ----------
        unit
            A unit value between 0 and 1.

        Returns
        -------
        A physical value, mapped from the unit value accoridng to the prior.
        """
        return super().value_for(unit, ignore_prior_limits=ignore_prior_limits)

    @property
    def parameter_string(self) -> str:
        return f"mean = {self.mean}, sigma = {self.sigma}"
